# Supplementary Material

# Estimating heritability via Haseman-Elston Regression 

setwd("") # insert the path of your working directory

# Loading data
load("dataset.RData")
# Description:
# The dataset contains information on 275 subjects, among which 212 come from Sardinian families (143 controls and 69 cases)
# Only 92 out of 143 unaffected subjects have genotyping data
# The remaining 63 unaffected subjects are retrieved from Sardinian population (not related with previous subjects)

#We keep the 212 related subjects. Subjects with FID>=9000 are the unrelated controls
data<-data[which(data$FID<9000),]


############ ############## ########### ############

# We show an example with protein Gc, then we show the loop to produce estimates for all proteins

# Sorting 
data_ord<-data[order(data$FID, data$IID, data$PAT, data$MAT),] # FID=family id, IID=individual id, PAT=father id, MAT=mother id
rownames(data_ord)<-data_ord$IID

# KINSHIP MATRIX COEFFICENTS
load("kinship_matrix.RData")
kmat<-kmat[as.character(data_ord$IID),as.character(data_ord$IID)]

# SHARED ENVIRONMENT MATRIX COEFFICENTS (coefficient=1 for each couple of siblings)
load("sharedenvironment_matrix.RData")
hmat<-hmat[as.character(data_ord$IID),as.character(data_ord$IID)]

# Source for Heritability estimation using Haseman-Elston regression:
# github: heritability_CIs 

source("Heritability_CIs.R")

# Standardized outcome variable (protein level)

# 1) Just with one protein

y<- data_ord$GC_HPA001526
y_ord<-as.matrix(scale(y))
n<- length(y)

# Covariates
mat <- list(kmat, hmat)
names(mat)<-c("Kinship", "SharedEnv")

intercept<-rep(cbind(1), 212)
intercept<- as.matrix(intercept)
intercept<- cbind(intercept,data_ord$MS, data_ord$SEX)

vc <- calc.vc(y_ord, intercept, mat, verbose = TRUE, max.iter = 20, eps = 1e-2)

# heritability 
hadd <- testVCcalcCIs(mat, "Kinship", vc$VC, vc$XtXinv, vc$var.resids, verbose = TRUE, CI.prob = 0.95, prob.eps = 0.01,  end.point.eps = 1e-3)
shared <- testVCcalcCIs(mat, "SharedEnv", vc$VC, vc$XtXinv, vc$var.resids, verbose = TRUE, CI.prob = 0.95, prob.eps = 0.01,  end.point.eps = 1e-3)

# --------------------------------------------   ---------------------------------   ------------------------------

# 2) With several proteins

# Heritability step 2: con kinship + shared environment matrix

# Covariates
mat <- list(kmat, hmat)
names(mat)<-c("Kinship", "SharedEnv")

intercept<-rep(cbind(1), 212)
intercept<- as.matrix(intercept)
intercept<- cbind(intercept,data_ord$MS, data_ord$SEX)

cov<-data_ord[8:63]  # cov is the ensamble of outcome variabkles (all the proteins)

her1  <- vector(mode = "list")
her2  <- vector(mode = "list")
prot  <- vector(mode = "list")
pvalue1  <- vector(mode = "list")
pvalue2  <- vector(mode = "list")
h2lower1  <- vector(mode = "list")
h2upper1 <- vector(mode = "list")
h2lower2  <- vector(mode = "list")
h2upper2 <- vector(mode = "list")

x=1

while(x<57){    # we have 56 proteins
  y_ord<- as.matrix(scale(cov[x]))

  vc <- calc.vc(y_ord, intercept, mat, verbose = TRUE, max.iter = 20, eps = 1e-2)
  
  
  # heritability and shared environment with 95% CIs
  
  hadd <- testVCcalcCIs(mat, "Kinship"  , vc$VC, vc$XtXinv, vc$var.resids, verbose = TRUE, CI.prob = 0.95, prob.eps = 0.01,  end.point.eps = 1e-3)
  shared <- testVCcalcCIs(mat, "SharedEnv"  , vc$VC, vc$XtXinv, vc$var.resids, verbose = TRUE, CI.prob = 0.95, prob.eps = 0.01,  end.point.eps = 1e-3)
  
  
  her1[x] <- hadd$propVar
  her2[x] <- shared$propVar
  
  pvalue1[x]  <- hadd$VCpval
  pvalue2[x]  <- shared$VCpval
  
  h2lower1[x]  <- hadd$VC_CI[[1]]
  h2upper1[x] <- hadd$VC_CI[[2]]
  
  h2lower2[x]  <- shared$VC_CI[[1]]
  h2upper2[x] <- shared$VC_CI[[2]]
  
  
  x<-x+1}

# Creating a dataframe for results:
DAT<-as.data.frame(names(cov))
names(DAT)[1]<-"Protein"
DAT$heritability<-as.numeric(her1)
DAT$Her_95LCL<-as.numeric(h2lower1)
DAT$Her_95UCL<-as.numeric(h2upper1)
DAT$Her_pvalue<-as.numeric(pvalue1)
DAT$adj_pvalue1<-p.adjust(as.numeric(pvalue1), method="BH")
DAT$shared_environment<-as.numeric(her2)
DAT$shared_environment_95LCL<-as.numeric(h2lower2)
DAT$shared_environment_95UCL<-as.numeric(h2upper2)
DAT$shared_environment_pvalue<-as.numeric(pvalue2)
DAT$adj_pvalue2<-p.adjust(as.numeric(pvalue2), method="BH")

write.table(DAT, sep=";", file="Heritability_results.csv", row.names = F, col.names = T, quote = F)


