# Supplementary Material

# Estimating proportion of protein level variability explained by significant SNPs + bootstrap BCA confidence interval


# forward.lmer: a function for stepwise regression using lmer mixed effects models
# Author: Rense Nieuwenhuis
forward.lmer <- function(
    start.model, blocks,
    max.iter=1, sig.level=FALSE,
    zt=FALSE, print.log=TRUE)
{
  
  # Initialysing internal variables
  log.step <- 0
  log.LL <- log.p <- log.block <- zt.temp <- log.zt <- NA
  model.basis <- start.model
  
  # Maximum number of iterations cannot exceed number of blocks
  if (max.iter > length(blocks)) max.iter <- length(blocks)
  
  # Setting up the outer loop
  for(i in 1:max.iter)
  {
    
    skip_to_next <- FALSE
    
    models <- list()
    
    tryCatch({
    # Iteratively updating the model with addition of one block of variable(s)
    # Also: extracting the loglikelihood of each estimated model
    for(j in 1:length(blocks))
    {
      
  
      models[[j]] <- update(model.basis, as.formula(paste(". ~ . + ", blocks[j])))
                   
      
    }
    
    LL <- unlist(lapply(models, logLik))
    
    # Ordering the models based on their loglikelihood.
    # Additional selection criteria apply
    for (j in order(LL, decreasing=TRUE))
    {
      
      
      ##############
      ############## Selection based significance of added variable-block
      ##############
      
      if(zt != FALSE)
      {
        b.model <- summary(models[[j]])$coefficients
        diff.par <- setdiff(rownames(b.model), rownames(summary(model.basis)$coefficients))
        if (length(diff.par)==0) break
        sig.par <- FALSE
        
        for (k in 1:length(diff.par))
        {
          if(abs(b.model[which(rownames(b.model)==diff.par[k]),3]) > zt)
          {
            sig.par <- TRUE
            zt.temp <- b.model[which(rownames(b.model)==diff.par[k]),3]
            break
          }
        }
        
        if(sig.par==TRUE)
        {
          model.basis <- models[[j]]
          
          # Writing the logs
          log.step <- log.step + 1
          log.block[log.step] <- blocks[j]
          log.LL[log.step] <- as.numeric(logLik(model.basis))
          log.zt[log.step] <- zt.temp
          blocks <- blocks[-j]
          
          break
        }
      }
    }}, error = function(e) { skip_to_next <<- TRUE})
    if(skip_to_next) { next } 
  }
  
  ## Create and print log
  log.df <- data.frame(log.step=1:log.step, log.block, log.LL, log.p, log.zt)
  if(print.log == TRUE) print(log.df, digits=4)
  
  ## Return the 'best' fitting model
  return(model.basis)
}
library(lme4qtl)
library(lme4)
library(readxl)

"%nin%"<-Negate("%in%")

setwd("") # insert the path of your working directory

#Loading Data
load("dataset.RData")
# We keep only the 155 unaffected controls with genotype data
data<-data[which(data$MS==0 & !is.na(data$rs1815606)),]

#Creating Mother label to consider siblings shared environment effect
data$MAT2<-data$MAT
data$MAT2[which(data$MAT!=0)]<- as.character(data$MAT)[which(data$MAT!=0)]
data$MAT2[which(data$MAT==0)]<- paste(as.character(data$IID)[which(data$MAT==0)], as.character(data$SUBFAMID)[which(data$MAT==0)], sep = "")


#Loading Kinship matrix
load("kinship_matrix.RData")
kmat<-kmat[as.character(data$IID),as.character(data$IID)]


#Protein of interest and SNPs with top 50 associations
name<-"GC_HPA001526"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 1)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]


name<-"PLAT_HPA003412"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 2)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]


name<-"ANXA1_HPA011271"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 3)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]

name<-"SOD1_HPA001401"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 4)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]

name<-"IRF8_HPA002531"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 5)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]

name<-"PTGER4_HPA012756"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 6)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]

name<-"FADD_HPA001464"
tab <- read_excel("Supplementary Table S3.xlsx", sheet = 7)
tab<-tab[order(tab$`p-value`),]
snps<-tab$`Immunochip SNP`[1:50]

# The code below is adapted for GC_HPA001526 protein, it should be updated where specified if protein is changed

snppp<-paste(snps, collapse = "+")

# Impute missing SNPs based on Hardy-weinberg equilibrium (very rare)
which(is.na(data[,snps]))
for(i in 1:50){
p<-(sum(data[,snps[i]], na.rm=T)/(length(which(!is.na(data[,snps[i]])==TRUE))*2))
data[is.na(data[,snps[i]]), snps[i]] <- sample(c(0, 1, 2), size = 1, replace = TRUE, prob = c((1-p)^2, (p*(1-p)), p^2))
}
which(is.na(data[,snps]))

#Standardizing protein levels
data[,as.character(name)] <-scale(data[,as.character(name)])

#Null model to start stepwise procedure
model2 <- as.formula(paste(name, " ~ (1|IID) + (1|MAT2) + SEX", collapse = " "))
m1 <- relmatLmer(model2, data=data,REML=F, relmat = list(IID = kmat), control = lmerControl(optimizer ="Nelder_Mead"))

#Stepwise procedure which includes SNPs at 0.0001 level
m1<-suppressMessages(summary(forward.lmer(m1, blocks=snps, max.iter=50, zt=3.719016)))

#p-values and significant SNPs
pvalues<-signif(1 - pchisq((m1$coefficients[,1]/m1$coefficients[,2])^2, 1), 10)
significant_snps<-names(pvalues[which(names(pvalues) %nin% c("(Intercept)", "SEX"))])
significant_snps2<-paste(significant_snps, collapse = "+")

#extract coefficients
beta<-m1$coefficients[,1]

#Estimating marginal proportion of protein level variability explained by significant SNPs

#Sex component
XBsex<-  beta[2] * data$SEX

#Significant SNPs (to be updated if protein is changed)
which(names(pvalues) %nin% c("(Intercept)", "SEX"))
sig_X1<-beta[3] * data[,as.character(names(beta)[3])]+
  beta[4] * data[,as.character(names(beta)[4])]+
  beta[5] * data[,as.character(names(beta)[5])]+
  beta[6] * data[,as.character(names(beta)[6])]
  
#All covariates component
XB<-sig_X1+XBsex

#Significant SNPs and other covariates covariance
cov<-cov(sig_X1
         , XBsex)

#heritability component
a<-as.data.frame(m1[["varcor"]][["IID"]])
#shared environment component
b<-as.data.frame(m1[["varcor"]][["MAT2"]])

#Marginal proportion of protein level variability explained by significant SNPs
R2SNPs<-(var(sig_X1)+cov) /( m1$sigma^2 + var(XB) + a$`(Intercept)`+b$`(Intercept)`)


# 95% Confidence interval using bias-corrected accelerated method

BCA2<-function(boot){
  #BCA method
  #Desired quantiles
  alpha<-0.05
  u <- c(alpha/2, 1-alpha/2) 
  
  #Compute constants
  z0 <- qnorm(mean(boot <= R2SNPs))
  zu <- qnorm(u)
  
  group_unique <- unique(data$SUBFAMID)
  
  n<-length(group_unique)
  I <- rep(NA, n)
  jack <- rep(NA, n)
  i<-1
  for(i in 1:n){
    #Remove ith household family
    xnew <- data[-which(data$SUBFAMID==group_unique[i]),]
    kmat3<-kmat[as.character(xnew$IID), as.character(xnew$IID)]
    
    #Estimate theta
    
    model2 <- as.formula(paste(name, " ~ (1|IID) + (1|MAT2) + SEX+",significant_snps2, collapse = " "))
    
    
    m1 <- relmatLmer(model2, data=xnew,REML=F, relmat = list(IID = kmat3), control = lmerControl(optimizer ="Nelder_Mead"))
    m1<-summary(m1)
    beta<-m1$coefficients[,1]
    
    #Estimating marginal proportion of protein level variability explained by significant SNPs
    
    #Sex component
    XBsex<-  beta[2] * xnew$SEX
    
    #Significant SNPs (to be updated if protein is changed)
    sig_X1<-beta[3] * xnew[,as.character(names(beta)[3])]+
      beta[4] * xnew[,as.character(names(beta)[4])]+
      beta[5] * xnew[,as.character(names(beta)[5])]+
      beta[6] * xnew[,as.character(names(beta)[6])]
    
    #All covariates component
    XB<-sig_X1+XBsex
    
    #Significant SNPs and other covariates covariance
    cov<-cov(sig_X1
             , XBsex)
    
    #heritability component
    a<-as.data.frame(m1[["varcor"]][["IID"]])
    #shared environment component
    b<-as.data.frame(m1[["varcor"]][["MAT2"]])
    VXF<-(var(sig_X1)+cov) /( m1$sigma^2 + var(XB) + a$`(Intercept)`+b$`(Intercept)`)
    
    I[i] <- (n-1)*(R2SNPs - VXF)
    jack[i]<-VXF
  }
  #Estimate a
  a <- (sum(I^3)/sum(I^2)^1.5)/6
  
  #Adjusted quantiles
  u_adjusted <- pnorm(z0 + (z0+zu)/(1-a*(z0+zu))) 
  
  #Accelerated Bootstrap CI
  Bca<-quantile(boot, u_adjusted)
  return(Bca)
}

# Bootstrap
"nendsWith"<-Negate("endsWith")
substrRight <- function(x, n){
  substr(x, nchar(x)-n+1, nchar(x))
}
VXFboot<- vector(mode = "list")
b<-data
b<-unique(b$SUBFAMID)
n<-length(b)
#Number of bootstrap replications
B<-1000
Esti<-numeric(B)
set.seed(1059)
for(i in 1:1000){
  index <-sample(b,
                 size=n,
                 replace=TRUE) 
  index<-as.data.frame(index)
  colnames(index)<-"SUBFAMID"
  
  tot3<-merge(data, index, by="SUBFAMID", all.y=T, all.x=F)
  kmat2<-kmat[as.character(tot3$IID),as.character(tot3$IID)]
  
  
  c1.rle <- rle(tot3$IID)
  tot3$id2 <- paste0(rep(c1.rle$values, times = c1.rle$lengths), "_",
                     unlist(lapply(c1.rle$lengths, seq_len)))
  
  colnames(kmat2)<-tot3$id2
  rownames(kmat2)<-tot3$id2
  
  kmat2[kmat2 ==1]<-0
  diag(kmat2)<-1
  
  kmat2[which(endsWith(rownames(kmat2),"_1")),which(nendsWith(colnames(kmat2), "_1"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_2")),which(nendsWith(colnames(kmat2), "_2"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_3")),which(nendsWith(colnames(kmat2), "_3"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_4")),which(nendsWith(colnames(kmat2), "_4"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_5")),which(nendsWith(colnames(kmat2), "_5"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_6")),which(nendsWith(colnames(kmat2), "_6"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_7")),which(nendsWith(colnames(kmat2), "_7"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_8")),which(nendsWith(colnames(kmat2), "_8"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_9")),which(nendsWith(colnames(kmat2), "_9"))]<-0
  kmat2[which(endsWith(rownames(kmat2),"_10")),which(nendsWith(colnames(kmat2), "_10"))]<-0
  
  tot3$MAT3<- paste(as.character(tot3$MAT2), substrRight(tot3$id2, 2), sep = "")

  model2 <- as.formula(paste(name, " ~ (1|id2) + (1|MAT3) + SEX+",significant_snps2, collapse = " "))
  
  
  skip_to_next <- FALSE
  
  tryCatch(  m1 <- suppressMessages(relmatLmer(model2, data=tot3,REML=F, relmat = list(id2 = kmat2), control = lmerControl(optimizer ="Nelder_Mead")))
, error = function(e) { skip_to_next <<- TRUE})

  if(skip_to_next) { next } 
  
  m1<-summary(m1)
  beta<-m1$coefficients[,1]
  
  #Estimating marginal proportion of protein level variability explained by significant SNPs
  
  #Sex component 
  XBsex<-  beta[2] * tot3$SEX
  
  #Significant SNPs (to be updated if protein is changed)
  sig_X1<-beta[3] * tot3[,as.character(names(beta)[3])]+
    beta[4] * tot3[,as.character(names(beta)[4])]+
    beta[5] * tot3[,as.character(names(beta)[5])]+
    beta[6] * tot3[,as.character(names(beta)[6])]

  #All covariates component
  XB<-sig_X1+XBsex
  
  #Significant SNPs and sex covariance
  cov<-cov(sig_X1
           , XBsex)
  
  a<-as.data.frame(m1[["varcor"]][["id2"]])
  a2<-as.data.frame(m1[["varcor"]][["MAT3"]])
  
  VXF<-(var(sig_X1)+cov) /( m1$sigma^2 + var(XB) + a$`(Intercept)`+a2$`(Intercept)`)
  
  
  VXFboot[i]<-VXF
  print(i)
}
boot<-unlist(VXFboot)
R2BcaCI<-suppressMessages(BCA2(boot))

#R2SNPs estimate
R2SNPs

#95% Bca CI
R2BcaCI
